#!/usr/bin/perl

##############################################################################
# $URL: https://svn.home.lan/winlib/trunk/code/version.pl $
# $Id: version.pl 48 2013-04-27 16:24:39Z jcurl $
##############################################################################
# File: version.pl
#
# Overview
#  Based on the current revision control system, try to get information about
#  the software package version.
#
# Requirements
#
# Description
#  We open a file called 'version.ini' that contains information about our
#  project. We use this to determine if this is a labelled release, or if
#  this is a snapshot build. The format of this file depends on the type of
#  revision control system that is in use.
#
#  CVS (Concurrent Versioning System)
#  ----------------------------------
#    [release]
#    version=1.0.0
#    label.cvs=WSL_CODE_0-0-0
#    
#    [snapshot]
#    version=1.1.0.%r_%d
#
#    [build]
#    build.cvs=<n>
#
#  For the CVS implementation, we check the current repository. If all the
#  files belong to the label '[Release].label' then we use the version
#  provided in '[Release].version. Else we use the version given in
#  '[Snapshot].version'. If we determine we're using a release version, we
#  use the build number given by '[build].build.cvs', else we increment
#  the build and store the new build number in 'version.ini' as the element
#  '[build].build.cvs'.
#
#  SVN (Subversion)
#  ----------------
#    [Release]
#    version=1.0.0
#    label.svn=<n>
# 
#    [Snapshot]
#    version=1.0.1.%r_%d
#
#  For the SVN implementation the same information exists for the CVS case,
#  except that the label is the SVN revision number.
#
#  For the build process, we maintain a separate file called 'version.txt'.
#  It maintains information about the build. Do NOT put this file under
#  revision control. It is maintained by the scripts and is only used when
#  you create a distribution from your build thereafter. All useful info
#  is obtained automatically from other files or from the revision control
#  system for you.
#
#  version.txt
#  -----------
#    [build]
#    build=<n>
#    rcs=<rcs>
#    version=<version>
#    note=<note>
#
#  These entries are created as necessary by this script and you don't need to
#  define these, or modify these. The 'build' is the number that this script
#  maintains, either by itself, or obtains with the help of the version
#  control system in use. The 'rcs' is the revision control system being used
#  and is a cache in case that the code is exported, we can still maintain if
#  this is a snapshot or a release. 'version' is the resulting version as
#  determined by this script. 'note' contains any information that might be
#  useful in how we got the version.
#
#  Command Line
#  ============
#  There are a number of options you can pass to this script to get version
#  information.
#
#  --release: Indicates this is for a release build, so if a connection must
#      made to a remote repository, do it (e.g. CVS)
#  --build: Get the build number. This depends on the revision control system.
#      * This is obtained from the file 'version.txt'.
#  --rcs: Get the revision control system in use. If we working in a revision
#      control system, we get this immediately, else we use the information
#      in 'version.ini'. This is to cater for snapshot releases which are
#      exported from the revision control system.
#      * CVS
#      * SVN
#  --note: Get the reason about how we got the version that we did.
#      * This is obtained from the file 'version.txt'
#
#  You must also give the path to where the file 'version.ini' in your code
#  repository is kept.
#
# Notes:
#
# To Do:
#  - The functions GetNextRCSVersion() both have some common sections that
#    should be factorised and put into a general purpose function.
#  - Expand SVN functionality to check URLs or paths for labels, instead of a
#    revision number. This can only be done against a "live" network though.
#
##############################################################################
use strict;
use warnings;

use Cwd;

use File::Spec::Functions qw( catfile path );

##############################################################################
# Global Variables for the program
my $cfgfile = "version.ini";
my $bldfile = "version.txt";
my $rcstype = undef;
my $checkfs = 0;

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetINIItem($$$$)
{
    my $file = shift;
    my $header = shift;
    my $searchkey = shift;
    my $defkey = shift;
    my ($headerfound, $value, $key);

    my @inifile;
    my $line;
    my $lineno;

    open(INIFILE, "< " . $file) || return $defkey;
    while (<INIFILE>) {
        chomp;
        push @inifile, $_;
    }
    close(INIFILE);

    $headerfound = 0;
    for $lineno (0 .. $#inifile) {
        $line = $inifile[$lineno];
        if ($line =~ m/^\s*\;/) {
            # Do nothing, it's a comment
        } elsif ($line =~ m/^\s*\[\s*(.*\S)\s*\]\s*$/) {
            # It's a section header. Check if it's the one we want
            if ($1 eq $header) {
                $headerfound = 1;
            } else {
                if ($headerfound == 1) {
                    return $defkey;
                }
            }
        } else {
            if ($headerfound == 1) {
                if ($line =~ m/^\s*(.*\S)\s*\=\s*(.*\S)\s*$/) {
                    $key = $1;
                    $value = $2;
                    if ($value =~ m/\"(.*)\"/) {
                        $value = $1;
                    }
                    if ($key eq $searchkey) {
                        return $value;
                    }
                }
            }
        }
    }
    return $defkey;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#  - Fix the case where 'value' is 'undef' and we should then delete the entry
#
##############################################################################
sub SetINIItem($$$$)
{
    my $file = shift;
    my $header = shift;
    my $searchkey = shift;
    my $value = shift;

    my @inifile;
    my $line;
    my $lineno;
    my ($added, $write);
    my ($headerfound, $key);

    if (open(INIFILE, "< $file")) {
        while (<INIFILE>) {
            chomp;
            push @inifile, $_;
        }
        close(INIFILE);
    }

    open(INIFILE, "> $file") || return undef;
    $headerfound = 0; $added = 0;
    for $lineno (0 .. $#inifile) {
        $write = 1;
        $line = $inifile[$lineno];
        if ($line =~ m/^\s*\;/) {
            # Do nothing, it's a comment
        } elsif ($line =~ m/^\s*\[\s*(.*\S)\s*\]\s*$/) {
            # It's a section header. Check if it's the one we want
            if ($1 eq $header) {
                $headerfound = 1;
            } else {
                # It's a different header
                if ($headerfound == 1) {
                    # We didn't find the key, so add the new key to the
                    # file now, and then dump the line.
                    unless ($added) {
                        print INIFILE "$searchkey=$value\n";
                        $added = 1;
                    }
                    $write = 0;
                }
            }
        } else {
            if ($headerfound == 1) {
                if ($line =~ m/^\s*(.*\S)\s*\=\s*(.*\S)\s*$/) {
                    $key = $1;
                    if ($key eq $searchkey) {
                        unless ($added) {
                            print INIFILE "$searchkey=$value\n";
                            $added = 1;
                        }
                        $write = 0;
                    }
                }
            }
        }
        if ($write) {
            print INIFILE "$line\n";
        }
    }
    unless ($added) {
        if ($headerfound) {
            print INIFILE "$searchkey=$value\n";
        } else {
            print INIFILE "[$header]\n";
            print INIFILE "$searchkey=$value\n";
        }
    }
    close(INIFILE);
}

##############################################################################
# Function: CheckInPath
#
# Inputs:
#  $file - The executable to test for in PATH
#
# Description:
#  Given an executable, look in the current system path to see if it exists.
#  We don't run the executable, or check that everything is installed (e.g.
#  if all relevant binaries are there). Just a quick rudimentary check that
#  works 99% of the time.
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub CheckInPath($)
{
    my $name = shift;
    my @path = path;
    my @pathext;

    my $pathentry;
    my $extentry;
    my $file;

    if (defined($ENV{PATHEXT})) {
        @pathext = split(/;/, $ENV{PATHEXT});
    } else {
        @pathext = ("");
    }

    foreach $pathentry (@path) {
        foreach $extentry (@pathext) {
            $file = File::Spec->catfile($pathentry, $name);
            if ($extentry ne "") {
                $file = $file . "." . $extentry;
            }
            print $file . "\n";
            if (-f $file) {
                return $file;
            }
        }
    }

    return undef;
}

##############################################################################
# Function: GetRevisionControlSystem
#
# Inputs:
#  $path - The path to the project
#
# Description:
#  Given a path, look for telltale signs of what revision control system
#  we're using.
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetRevisionControlSystem($)
{
    my $oldpath = getcwd;
    my $path = shift;

    # Check if we're using CVS and the binaries exist on the system
    if (-f "$path/CVS/Root") {
        if (defined(CheckInPath("cvs"))) {
            return "cvs";
        }
    }

    # Check if we're using SVN and the binaries exist on the system
    # Works with SVN 1.6, but doesn't work with SVN 1.7
    #if (-f "$path/.svn/entries") {
    #    if (defined(CheckInPath("svn"))) {
    #        return "svn";
    #    }
    #}
    chdir $path;
    unless (open(SVN, "svn info 2>/dev/null |")) {
        chdir $oldpath;
        return 0;
    }

    while(<SVN>) {
        chomp;
        if ($_ =~ m/^Revision: (\d+)$/) {
            chdir $oldpath;
            return "svn";
        }
    }

    chdir $oldpath;
    return undef;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetBuildNumber($)
{
    my $path = shift;
    my $build;

    # We return the build number in the version.ini file but don't increment
    $build = GetINIItem("./$bldfile", "build", "build", "0");
    return $build;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetDate()
{
    my @timeData;
    my ($year, $month, $day);
    my $date;

    @timeData = localtime();
    $year = $timeData[5] + 1900;
    $month = $timeData[4] + 1;
    $day = $timeData[3];

    $date = sprintf ("%04d%02d%02d", $year, $month, $day);
    return $date;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub ExpandVersionString($$)
{
    my $inver = shift;
    my $build = shift;
    my $version;
    my $date;

    $version = $inver;
    $date = GetDate();

    $version =~ s/%%/%/g;
    $version =~ s/%b/$build/g;
    $version =~ s/%r/$rcstype/g;
    $version =~ s/%d/$date/g;

    return $version;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetSVNVersion($)
{
    my $path = shift;
    my $build;
    my $oldpath;

    # We return the 'svnversion' build
    $oldpath = getcwd;
    chdir $path;
    unless (open(BUILD, "svnversion |")) {
        chdir $oldpath;
        return 0;
    }
    
    $build = <BUILD>;
    chomp $build;

    close(BUILD);
    chdir $oldpath;

    return $build;
}

sub GetSVNBuildNumber($)
{
    my $path = shift;
    my $build;
    my $svnver;

    # We get the subversion number and convert it to an integer
    $svnver = GetSVNVersion($path);
    if ($svnver =~ m/(\d+:)?(\d+)M?S?P?/) {
        $build = $2;
    } else {
        $build = $svnver;
    }
    return $build;
}

sub GetSVNURL($)
{
    my $path = shift;
    my $url;
    my $oldpath;

    $oldpath = getcwd;
    chdir $path;
    unless (open(INFO, "svn info |")) {
        return undef;
    }
    
    while (<INFO>) {
        chomp;
        if ($_ =~ m/^URL: (.+)$/) {
            $url = $1;
        }
    }

    close(INFO);
    chdir $oldpath;

    return $url;
}

sub GetSVNRepoRoot($)
{
    my $path = shift;
    my $url;
    my $oldpath;

    $oldpath = getcwd;
    chdir $path;
    unless (open(INFO, "svn info |")) {
        return undef;
    }
    
    while (<INFO>) {
        chomp;
        if ($_ =~ m/^Repository Root: (.+)$/) {
            $url = $1;
        }
    }

    close(INFO);
    chdir $oldpath;

    return $url;
}

sub GetSVNDiff($$)
{
    my $patha = shift;
    my $pathb = shift;
    my $lines = 0;

    # Redirect STDERR also to stdout, in case there's an error
    unless (open(DIFF, "svn diff $patha $pathb 2>&1 |")) {
        return undef;
    }

    while (<DIFF>) {
        chomp;
        $lines++;
    }

    close(DIFF);
    return $lines;
}
        

sub GetNextSVNVersion($)
{
    # We compare the element [release].label.svn with 'svnversion'. If they
    # match, we return the expanded form of [release].version. Else we return
    # the expanded form of [snapshot].version.
    my $path = shift;
    my $svnlive;
    my $laboff;
    my $labcur;
    my $build;
    my $version;
    my $release;
    my $expversion;
    my $note;
    my ($patha, $pathb, $root, $url);
    my $diffs;

    $laboff = GetINIItem("$path/$cfgfile", "release", "label.svn", undef);

    # Check if we're comparing a revision number, or a label from the root
    # of the repository
    if ($laboff =~ m/^(\d+)$/) {
        # This is a version number
        $labcur = GetSVNVersion($path);

        # Get the version number to use
        if ($laboff eq $labcur) {
            $version = GetINIItem("$path/$cfgfile", "release", "version", undef);
            $note = "SVN Release";
        } else {
            $version = GetINIItem("$path/$cfgfile", "snapshot", "version", undef);
            $note = "SVN Wrong Version/Modified Files";
        }
        $expversion = ExpandVersionString($version, $labcur);
        $build = GetSVNBuildNumber($path);
    } else {
        $svnlive = $checkfs;
        if ($svnlive) {
            $note="";
            # We assume this is a path from the root of the repository

            # Check the format of the string
            #   http://... 
            #     this is an absolute path
            #   /path/to/folder
            #      This is a reference from the base path. Run 'svn info' and parse
            #     the response for 'Repository Root: %s' and use that as the base
            #     appending the path appropriately.
            #
            #      Then we need to compare them with 'svn diff a b', where a is the
            #     field from 'svn info' being 'URL: %s' and b is the tag. If there are
            #     differences, we know that it's changed since the label.
            if ($laboff =~ m|^[A-Za-z]+://(.+)$|) {
                # This is an absolute URL, so we use that
                $pathb = $laboff;
            } elsif ($laboff =~ m|^/(.+)$|) {
                # We assume this is a relative path
                $root = GetSVNRepoRoot($path);
                if (!defined($root)) {
                    $note = "SVN Root Repository unknown";
                    $release = 0;
                    $svnlive = 0;
                } else {
                    $pathb = "${root}${laboff}";
                }
            } else {
                # Or this type of relative path
                $root = GetSVNRepoRoot($path);
                if (!defined($root)) {
                    $note = "SVN Root Repository unknown";
                    $release = 0;
                    $svnlive = 0;
                } else {
                    $pathb = "${root}/${laboff}";
                }
            }

            if ($svnlive) {
                # Find the path that we compare to (our current repo)
                $patha = GetSVNURL($path);
                if (!defined($patha)) {
                    $note = "SVN Root URL unknown";
                    $release = 0;
                    $svnlive = 0;
                }
            }

            if ($svnlive) {
                # Actually perform the diff to see if everything is the same
                $diffs = GetSVNDiff($patha, $pathb);
                if ($diffs == 0) {
                    $note = "SVN Release";
                    $release = 1;
                } else {
                    $note = "SVN tag not matched";
                    $release = 0;
                }
            }
            
        } else {
            $note = "SVN connection ignored";
            $release = 0;
        }

        if ($release) {
            # This is a release
            $build = int(GetSVNBuildNumber($path));
            $version = GetINIItem("$path/$cfgfile", "release", "version", undef);
            $expversion = ExpandVersionString($version, $build);
        } else {
            # This is a snapshot. We don't add one to the build like in CVS.
            $build = int(GetSVNBuildNumber($path));
            $version = GetINIItem("$path/$cfgfile", "snapshot", "version", undef);
            $expversion = ExpandVersionString($version, $build);
        }
    }

    # Write the build number to the script (in case we package it now and
    # and release it, we can use this information).
    SetINIItem("./$bldfile", "build", "build", $build);
    SetINIItem("./$bldfile", "build", "rcs", $rcstype);
    SetINIItem("./$bldfile", "build", "version", $expversion);
    SetINIItem("./$bldfile", "build", "note", $note);

    return $expversion;
}

##############################################################################
# Function: GetNextCVSVersion
#
# Inputs:
#
# Description:
#  Apply a set of rules to determine if this is a released version or a
#  snapshot version.
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetCVSBuildNumber($)
{
    my $path = shift;
    my $build;

    # We return the build number in the version.ini file but don't increment
    $build = GetINIItem("$path/$cfgfile", "build", "build.cvs", "0");
    return $build;
}

sub GetNextCVSVersion($)
{
    my $path = shift;
    my $oldpath;
    my ($version, $expversion, $build, $note);
    my $cvslive;
    my $cvserr;
    my $release;
    my ($cvsdiff, $cvstag, $cvsm);
    my $label;

    # If we're not allowed to check the file system, we assume we're using
    # a snapshot release. This is because checking differences with CVS can
    # be a time consuming operation, or the server may not be online. So by
    # default we should probably assume that we don't have a connection to
    # the server.
    $cvslive = $checkfs;

    if ($cvslive) {
        $oldpath = getcwd;
        chdir $path;

        # Check that we have a connection to the CVS server
        unless (open (CVSLIVE, "cvs version 2> /dev/stdout |")) {
            # Problem executing CVS
            $note = "CVS command error";
            $cvslive = 0;
            $release = 0;
        } else {
            $release = 1; $cvserr = 0;
            while (<CVSLIVE>) {
                if ($_ =~ m/error/i) {
                    $release = 0;
                    $cvserr = 1;
                }
            }
            close (CVSLIVE);
            if ($cvserr) {
                $note = "CVS connection error";
                $cvslive = 0;
            }
        }

        # If we have a connection, check if the repository has changed or not
        if ($cvslive) {
            $cvstag = 0;
            $label = GetINIItem("$path/$cfgfile", "release", "label.cvs", "x");
            unless (open (CVSDIFF, "cvs diff -R -r $label --brief 2> $path/out.txt | grep differ | wc -l |")) {
                $note = "cvs diff command error";
                $cvslive = 0;
                $release = 0;
            } else {
                chomp($cvsdiff = <CVSDIFF>);
                close(CVSDIFF);
                if (open (CVSDIFF, "< $path/out.txt")) {
                    while (<CVSDIFF>) {
                        chomp;
                        if ($_ =~ m/aborted/) {
                            $cvstag++;
                        }
                        if ($_ =~ m/is not in file/) {
                            $cvstag++;
                        }
                    }
                    close (CVSDIFF);
                    unlink "$path/out.txt";
                }
            }
        }

        if ($cvslive) {
            unless (open (CVSM, "cvs -qn update 2> $path/out.txt | grep -G \"[MAR] \" | wc -l |")) {
                $note = "cvs update command error";
                $cvslive = 0;
                $release = 0;
            } else {
                chomp($cvsm = <CVSM>);
                close(CVSM);
                if (open (CVSM, "< $path/out.txt")) {
                    while (<CVSM>) {
                        chomp;
                        if ($_ =~ m/lost/) {
                            $cvstag++
                        }
                    }
                    close(CVSM);
                    unlink "$path/out.txt";
                }
            }
        }

        if ($cvslive) {
            if ($cvstag == 0 && $cvsm == 0) {
                if ($cvsdiff == 0) {
                    $release = 1;
                    $note = "CVS Release";
                } else {
                    $release = 0;
                    $note = "CVS Modified Files";
                }
            } else {
                if ($cvstag) {
                    $release = 0;
                    $note = "CVS new files, or tag missing";
                } else {
                    $release = 0;
                    $note = "CVS unchecked in files";
                }
            }
        }
        unlink "$path/out.txt";
        chdir $oldpath;
    } else {
        $note = "CVS connection ignored";
        $release = 0;
    }

    if ($release) {
        # This is a release
        $build = int(GetCVSBuildNumber($path));
        $version = GetINIItem("$path/$cfgfile", "release", "version", undef);
        $expversion = ExpandVersionString($version, $build);
    } else {
        # This is a snapshot
        $build = int(GetCVSBuildNumber($path)) + 1;
        $version = GetINIItem("$path/$cfgfile", "snapshot", "version", undef);
        $expversion = ExpandVersionString($version, $build);

        # This element is only written to if we're not using a release. That
        # is, we are using CVS but the build number needed to be incremented.
        SetINIItem("$path/$cfgfile", "build", "build.cvs", $build);
    }

    # Write the build number to the script (in case we package it now and
    # and release it, we can use this information).
    SetINIItem("./$bldfile", "build", "build", $build);
    SetINIItem("./$bldfile", "build", "rcs", $rcstype);
    SetINIItem("./$bldfile", "build", "version", $expversion);
    SetINIItem("./$bldfile", "build", "note", $note);

    return $expversion;
}

##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetNextGENVersion($)
{
    # Because we're not using a revision control system, we have to rely on
    # the 'version.ini' file that is distributed
    my $path = shift;
    my $version;

    $version = GetINIItem("./$bldfile", "build", "version", "");
    return $version;
}

##############################################################################
# Function: GetNextVersion
#
# Inputs:
#
# Description:
#  Based on the RCS in use, determine the next version (call the appropriate
#  function to do this for us)
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
sub GetNextVersion($)
{
    my $path = shift;

    if (defined($rcstype)) {
        if ($rcstype eq "svn") {
            return GetNextSVNVersion($path);
        } elsif ($rcstype eq "cvs") {
            return GetNextCVSVersion($path);
        }
    }

    return GetNextGENVersion($path);
}

##############################################################################
##############################################################################
# Function: 
#
# Inputs:
#
# Description:
#
# Design:
#
# Results:
#
# Errors:
#
# Notes:
#
# To Do:
#
##############################################################################
my $path;
my $getbuildnum = 0;
my $getrcstype = 0;
my $getnote = 0;
my $version;
my $note;

while ($#ARGV >= 0) {
    if ($ARGV[0] eq "--release") {
        $checkfs = 1;
    } elsif ($ARGV[0] eq "--build") {
        $getbuildnum = 1;
    } elsif ($ARGV[0] eq "--rcs") {
        $getrcstype = 1;
    } elsif ($ARGV[0] eq "--note") {
        $getnote = 1;
    } elsif ($ARGV[0] ne "") {
        if (defined($path)) {
            die "version.pl [--release] [--build] [--rcs] [--note] path\n";
        } else {
            $path = $ARGV[0];
        }
    }
    shift @ARGV;
}

unless (defined($path)) {
    $path = getcwd;
}

$rcstype = GetRevisionControlSystem($path);

##############################################################################
# Check if we should just report the current build number
if ($getbuildnum) {
    print GetBuildNumber($path) . "\n";
    exit 0;
}

##############################################################################
# Check if we should just report the revision control system detected
if ($getrcstype) {
    if (defined($rcstype)) {
        print "$rcstype\n";
    } else {
        print "Dist\n";
    }
    exit 0;
}

##############################################################################
# Check if we get the notes about the build
if ($getnote) {
    $note = GetINIItem("./$bldfile", "build", "note", "");
    print "$note\n";
    exit 0;
}

##############################################################################
# Check the version and decide if we return the official version string
# or if we report the snapshot version string.
$version = GetNextVersion($path);
print "$version\n";
